@@ -1,18 +1,18 @@
 #!/usr/bin/python3
-import os
-import sys
-import apt
 import gettext
-import tempfile
+import os
 import subprocess
-import mintcommon
-import gi
+import sys
 import locale
 
+import gi
 gi.require_version('Gtk', '3.0')
 gi.require_version('Vte', '2.91')
 from gi.repository import Gtk, Vte, GLib
 
+import apt
+import mintcommon.aptdaemon
+
 # i18n
 APP = 'mintsources'
 LOCALE_DIR = "/usr/share/linuxmint/locale"
@@ -85,21 +85,49 @@ class Foreign_Browser():
             treeview.append_column(col)
             col.set_sort_column_id(PKG_REPO_VERSION)
 
-        self.apt = mintcommon.APT(self.window)
+        self.apt = mintcommon.aptdaemon.APT(self.window)
 
         cache = apt.Cache()
+
+        # python-apt doesn't give us a constant for this value
+        # it's "required" in English, but it could be anything in
+        # other languages
+        required_priority = cache['dpkg'].installed.priority
+
         for key in cache.keys():
             pkg = cache[key]
             if (pkg.is_installed):
                 candidate_version = pkg.candidate.version
                 installed_version = pkg.installed.version
 
-                if not pkg.candidate.downloadable:
-                    # The candidate is not downloadable...
-                    # See if there's a version that is...
+                if not self.downgrade_mode:
+                    # remove mode
+                    # Find packages which aren't downloadable
+                    if not pkg.candidate.downloadable:
+                        downloadable = False
+                        for version in pkg.versions:
+                            if version.downloadable:
+                                downloadable = True
+                        if not downloadable and pkg.essential == False and pkg.installed.priority != required_priority:
+                            iter = self.model.insert_before(None, None)
+                            self.model.set_value(iter, PKG_ID, "%s" % (pkg.name))
+                            self.model.set_value(iter, PKG_CHECKED, False)
+                            self.model.set_value(iter, PKG_NAME, "<b>%s</b>" % pkg.name)
+                            self.model.set_value(iter, PKG_INSTALLED_VERSION, installed_version)
+                            self.model.set_value(iter, PKG_REPO_VERSION, "")
+                            self.model.set_value(iter, PKG_SORT_NAME, "%s" % (pkg.name))
+                else:
+                    # downgrade mode
+                    # Find packages which candidate isn't available
                     best_version = None
-                    for version in pkg.versions:
-                        if version.downloadable:
+                    if not pkg.candidate.downloadable:
+                        # foreign packages which installed version doesn't exist in the repositories
+                        for version in pkg.versions:
+                            if not version.downloadable:
+                                continue
+                            if version.version == installed_version:
+                                continue
+
                             if best_version is None:
                                 best_version = version
                             else:
@@ -110,27 +138,29 @@ class Foreign_Browser():
                                     return_code = subprocess.call(["dpkg", "--compare-versions", version.version, "gt", best_version.version])
                                     if return_code == 0:
                                         best_version = version
+                    elif candidate_version == installed_version:
+                        # packages which installed version isn't the one available from the highest priority repo
+                        # typical example is an ubuntu version installed despite a inferior 700 version being available
+                        best_version = None
+                        for version in pkg.versions:
+                            if not version.downloadable:
+                                continue
+                            if version.version == installed_version:
+                                continue
+                            if version.policy_priority > 500:
+                                if best_version is None or version.policy_priority > best_version.policy_priority:
+                                    best_version = version
 
-                    if self.downgrade_mode:
-                        if best_version is not None:
-                            for origin in best_version.origins:
-                                iter = self.model.insert_before(None, None)
-                                self.model.set_value(iter, PKG_ID, "%s=%s" % (pkg.name, best_version.version))
-                                self.model.set_value(iter, PKG_CHECKED, False)
-                                self.model.set_value(iter, PKG_NAME, "<b>%s</b>" % pkg.name)
-                                self.model.set_value(iter, PKG_INSTALLED_VERSION, installed_version)
-                                self.model.set_value(iter, PKG_REPO_VERSION, "%s (%s)" % (best_version.version, origin.archive))
-                                self.model.set_value(iter, PKG_SORT_NAME, "%s %s" % (best_version.source_name, pkg.name))
-                                break
-                    else:
-                        if best_version is None:
+                    if best_version is not None:
+                        for origin in best_version.origins:
                             iter = self.model.insert_before(None, None)
-                            self.model.set_value(iter, PKG_ID, "%s" % (pkg.name))
+                            self.model.set_value(iter, PKG_ID, "%s=%s" % (pkg.name, best_version.version))
                             self.model.set_value(iter, PKG_CHECKED, False)
                             self.model.set_value(iter, PKG_NAME, "<b>%s</b>" % pkg.name)
                             self.model.set_value(iter, PKG_INSTALLED_VERSION, installed_version)
-                            self.model.set_value(iter, PKG_REPO_VERSION, "")
-                            self.model.set_value(iter, PKG_SORT_NAME, "%s" % (pkg.name))
+                            self.model.set_value(iter, PKG_REPO_VERSION, "%s (%s)" % (best_version.version, origin.archive))
+                            self.model.set_value(iter, PKG_SORT_NAME, "%s %s" % (best_version.source_name, pkg.name))
+                            break
 
         treeview.show()
         treeview.connect("row-activated", self.treeview_row_activated)
